# This file documents Guardian's use of the Ezca channel access
# object. In particular it is meant to document exactly how the Ezca
# object is used; how it's intialized, how it's methods are called,
# what return values it is expecting, etc.  Also a bit of the intended
# use of the object.
#
# Ezca's original purpose was as a wrapper of the EPICS pyepics Device
# class, for managing interaction with groups of common prefixed EPICS
# channels.
#
# http://cars9.uchicago.edu/software/python/pyepics3/devices.html
#
# Guardian user code would use the interface to read values from and
# write values to EPICS channels on the network, and the Ezca object
# (and ultimately the underlying pyepics.Device) would manage the
# connections to those channels.
#
# Currently guardian expects an 'ezca' module to exist, for it to
# provide an Ezca class, and for the Ezca class to provide certain
# methods and properties with certain return values.  However, the
# guardian user code is in no way required to use the interface or to
# even make any reference to it.  The entire module can in fact
# therefore be a no-op.  But it's important that at least guardian
# gets what it expects out of it, even if none of it is used.
#
# That said, any use of guardian may want to provide an Ezca module
# that is actually useful in the user code for the task at hand.
#
#
# MODULE
# ======
#
# The ezca module should provide the following objects, which are all
# imported by guardian:
#
#   from ezca import Ezca, EzcaError, EzcaConnectError
#
# PVS
# ===
#
# 'pv' objects are container objects that hold state of some process
# variable in use by the system.  In EPICS these hold state of some
# active channel access connection.  The LIGO Ezca object provides
# access to the underlying pvs via a dict-like get/setitem interface
# to the object:
#
# >>> prefix = 'T1:'
# >>> ezca = Ezca(prefix):
# >>> a = ezca['CHANNEL_A']
# >>> ezca['CHANNEL_B'] = 4
#
# In EPICS Ezca each pv maintains the state of a particular EPICS
# channel while guardian is active.  They are instantiated in the
# object when they are first accessed.
#
# Guardian assumes that Ezca has a 'pvs' property, assumed to be a
# dictionary of individual pv objects keyed by the names used to
# access them.  The only real requirement is that the 'pvs' property
# be a container object that supports the len() method.
#
# If a setpoint "snapshot" is requested (see SETPOINTS below), the
# guardian will actually iterate over the 'pvs' object to look at the
# value of each pv.  In that case each individual pv objects must have
# the following properties:
#
# pv.pvname:     string name of the pv
# pv.char_value: string representation of the value of the channel
#
#
# CONNECTIONS
# ===========
#
# Guardian wants to check the connection status of all pvs before
# executing state code.  If any of pvs have lost connection to their
# underlying channel, guardian will go into a connection error
# (CERROR) state and will forgo executing any state code until all
# connections have been reestablished.  Guardian does this by calling
# the ezca.check_connections() method.  This method should return an
# list of pv objects (see above) that are currently having connections
# issues.  Guardian will keep running check_connections() every cycle,
# not executing any state code, until it returns an empty list.
#
#  dead_channel_list = ezca.check_connections()
#
#
# SETPOINTS
# =========
#
# Like the connection checking described above, Guardian has the
# ability to provide information from setpoint checks performed by the
# Ezca object.  A setpoint is any value written by Ezca (usually via
# the constituent pvs).
#
# If the system.ca_channels variable is defined, guardian will pass it
# directly to the ezca.init_setpoints() method.  The typical usage
# would be for system.ca_channels to be a list of
# (channel_name, initial_setpoint_value) tuples.
#
#   ezca.init_setpoints(system.ca_channels)
#
# Setpoints should be stored in a ezca.setpoints property, which
# should be a dictionary keyed on channel name as PVS above, with
# (pvname, setpoint_value) tuples as values.
#
# If the system.ca_monitor variable is set to True, once a cycle
# guardian will execute the ezca.check_setpoints() method, which is
# expected to return a list of setpoint status tuples for channels
# whose value does not equal their setpoint.
#
#   modified_channel_list = ezca.check_setpoints()
#
# where modified_channel_list is a list of tuples:
#
#   (channel_name, setpoint_value, current_value, difference)
#
# Guardian can also request that the ezca object write all current
# setpoint values to a file to disk.  This is done with the following
# method:
#
#   ezca.setpoint_snap(path_to_snap_file)
#
#
# EXCEPTIONS
# ==========
#
# If the ezca object (or any user code for that matter) throws an
# EzcaError exception, a guardian ERROR error status will be reported.
#
# If the ezca object (or any user code for that matter) throws an
# EzcaConnectError exception, a guardian CERROR connection error
# status will be reported (see CONNECTIONS above).

############################################################
############################################################

# Variable holding version of library.  This is a string but it is
# expected to have a particular form of three integers separated by a
# period ('.') with 4, 8, and 8 bit values respectively.
#
# REQUIRED
# TYPE: str
# VALUE: three integers separated by '.', <4>.<8>.<8> bit
#        respectively.
__version__ = '0.0.0'


# Ezca object
#
# REQUIRED
# TYPE: object
class Ezca(object):
    # Guardian initializes in the worker subprocess only.  Ezca init
    # should accept the following keyword arguments:
    #
    #   'prefix': channel access prefix string as specified in the
    #   system module 'ca_prefix' variable.  It can be used or not
    #   used in the interface as appropriate.
    #
    #   'logger': standard python logging object that can be used to
    #   log activity in the object, such as channel read or writes.
    #   Using this provided object will prefix log writes with
    #   information about the state of the system, such as the time,
    #   node name, and current state.
    #
    def __init__(self, prefix='', logger=True):
        self._prefix = prefix
        self._log = logger

    # Channel prefix value.  This property should always return a
    # string, and if nothing else should return the 'prefix' string as
    # provided at initialization above.
    #
    # REQUIRED
    # TYPE: str
    @property
    def prefix(self):
        return self._prefix

    # Dictionary of pvs.  Keys are channel names, values are pv
    # objects with the following attributes:
    #
    #   pv.pvname: string
    #   pv.char_value: string of current value
    #
    # At very least len(ezca.pvs) must not fail.
    #
    # REQUIRED
    # TYPE: dict
    # VALUE: {channel_name: pv}
    @property
    def pvs(self):
        return {}

    # Method to check pv connection status.  Should return a list of
    # string channel names for channels with connection errors.
    #
    # REQUIRED
    # OUTPUT: list of str
    def check_connections(self):
        return []

    # Method to Initialize setpoints.  The value of system.ca_channels
    # will be passed directly to this method.
    #
    # CONDITIONAL: called if system.ca_channels is defined
    # INPUT: system.ca_channels defined
    def init_setpoints(self, table):
        pass

    # Dictionary of all current setpoints.  Keys are channel names
    # (see self.pvs), with (pvname, setpoint_value) tuples values.
    #
    # REQUIRED
    # TYPE: dict
    @property
    def setpoints(self):
        return {}

    # Method to request writing setpoint values to a file.
    #
    # CONDITIONAL: guardian system archive defined
    # INPUT: str path
    def setpoint_snap(self, snapfile):
        pass

    # Method to check current channel values against stored setpoints.
    # Should return a list of tuples of the form:
    #   (full_channel_name, setpoint_value, current_value, difference)
    #
    # CONDITIONAL: system.ca_monitor == True
    # INPUT: str path
    def check_setpoints(self):
        return []


# PV class provided as a convenience to show the properties expected
# in the ezca.pvs dictionary value objects.
class PV:
    # string name of pv
    @property
    def pvname(self):
        return ''

    # numeric/native value
    @property
    def value(self):
        return 0

    # string representation of value
    @property
    def char_value(self):
        return str(self.value)


# Exception that produces ERROR condition in guardian
class EzcaError(Exception):
    pass

# Exception that produces CERROR condition in guardian
class EzcaConnectError(Exception):
    pass



import pcaspy
import threading

from . import const
from .db import guarddb


# add the REQUEST_ENUM interface channel
guarddb['REQUEST_ENUM'] = {
    'type': 'enum',
    'enums': (),
    'writable': True,
    }


def request_enum_index(request):
    """Returns index of request string in REQUEST_ENUM"""
    return guarddb['REQUEST_ENUM']['enums'].index(str(request))


class CAServerError(Exception):
    pass

class CADriver(pcaspy.Driver):
    def __init__(self, request_event):
        super(CADriver, self).__init__()
        self._state_index = {}
        self._request_event = request_event

    ##########
    # these methods are for the external interface
    # NOTE: caller must updatePV after set

    def __getitem__(self, channel):
        value = self.getParam(channel)
        if guarddb[channel]['type'] == 'enum':
            if guarddb[channel].get('guardbool', False) and value == 0:
                value = False
            else:
                value = guarddb[channel]['enums'][value]
        return value

    def __setitem__(self, channel, value):
        # handle state readbacks
        if guarddb[channel].get('guardstate', False):
            nvalue = self._state_index[value]
            self.setParam(channel+'_S', value)
            self.setParam(channel+'_N', nvalue)
        # resolve enums, since setParam expects numeric value
        if guarddb[channel]['type'] == 'enum':
            value = guarddb[channel]['enums'].index(str(value))
        # update REQUEST_ENUM
        if channel == 'REQUEST' and value in guarddb['REQUEST_ENUM']['enums']:
            self.setParam('REQUEST_ENUM', request_enum_index(value))
        self.setParam(channel, value)

    ##########
    # these methods are for the internal interface, e.g. CA clients

    def write(self, channel, value):
        # NOTE: for enum records the value here is the numeric value,
        # not the string, and setParam() expects the numeric value.

        # reject writes to non-writable channels
        if not guarddb[channel].get('writable', False):
            return False

        # reject values that don't correspond to an actual index of
        # the enum
        # FIXME: this is apparently a feature? of cas that allows for
        # setting numeric values higher than the enum?
        if guarddb[channel]['type'] == 'enum' \
           and value >= len(guarddb[channel]['enums']):
            return False

        if channel in ['REQUEST', 'REQUEST_ENUM']:
            if channel == 'REQUEST':
                # reject invalid state requests
                if value not in self._state_index.keys():
                    return False
                if value in guarddb['REQUEST_ENUM']['enums']:
                    self.setParam('REQUEST_ENUM', request_enum_index(value))
                svalue = value
            elif channel == 'REQUEST_ENUM':
                self.setParam('REQUEST_ENUM', value)
                svalue = guarddb['REQUEST_ENUM']['enums'][value]

            self.setParam('REQUEST', svalue)
            self.setParam('REQUEST_S', svalue)
            self.setParam('REQUEST_N', self._state_index[svalue])
            self._request_event.set()

        elif channel == 'MODE':
            # value as string
            mode = guarddb['MODE']['enums'][value]
            # reject attempts to set MANAGED mode directly without
            # specifying manager
            if mode == 'MANAGED':
                return False
            self.setParam('MODE', value)
            # make sure manager channel is unset if switching away
            # from managed mode
            if mode in ['AUTO', 'MANUAL']:
                self.setParam('MANAGER', '')

        elif channel == 'MANAGER':
            # reject setting manager to empty string
            if value == '':
                return False
            self.setParam('MANAGER', value)
            self.setParam('MODE', guarddb['MODE']['enums'].index('MANAGED'))

        else:
            self.setParam(channel, value)

        self.updatePVs()

        return True


class CAServer(object):
    def __init__(self, system_name):
        self.prefix = const.CAS_PREFIX_FMT.format(IFO=const.IFO, SYSTEM=system_name)
        self._request_event = threading.Event()
        self._server = pcaspy.SimpleServer()
        #self._server.setDebugLevel(4)
        self._loaddb()
        self._driver = CADriver(self._request_event)
        # clear undefined alarms at startup
        for chan in guarddb:
            self._driver.setParamStatus(chan, pcaspy.Severity.NO_ALARM, pcaspy.Severity.NO_ALARM)
        self._driver.updatePVs()
        self._running = True
        self.daemon = True


    def _loaddb(self):
        self._server.createPV(self.prefix, guarddb)


    def process(self, timeout):
        """Server process select loop

        Returns in max `timeout` seconds, but usually much faster
        (this is presumably the timeout of the select itself).

        """
        self._server.process(timeout)

    ########################################

    def __getitem__(self, channel):
        """get PV value

         """
        return self._driver[channel]


    def __setitem__(self, channel, value):
        """set PV value

        ** Caller required to run updatePVs **

         """
        self._driver[channel] = value


    def updatePVs(self):
        """Push updates to driver

         """
        # The driver.updatePVs() call is the heaviest interaction with
        # the cas, consuming more processing time than anything else.
        # Calling it after every setitem call causes the load on the
        # system to be unacceptably high.
        #
        # FIXME: The placement of this also seems to have a strong
        # affect on the behavior of the EPICS interaction with
        # external clients.  One would assume that it only needs to be
        # called once per cycle of the main loop, to push all param
        # changes that happened in the loop.  But if that's done, the
        # EPICS interaction with the clients seems to become very odd,
        # and client puts don't necessarily show up as expected
        # (they're there in EPICS but not in pcaspy???  Why does this
        # seem to affect client writes??).  Needs more investigation.
        #
        # Anyway...  PLACEMENT OF THIS CALL NEEDS TO BE CAREFULLY
        # CONSIDERED AND TESTED.
        self._driver.updatePVs()


    def update_state_index(self, index, init=False):
        """Update state index dictionary

        """
        self._driver._state_index = index
        if init:
            return
        # update all guardstate channels, in case the state:index
        # mapping has changed
        for channel, entry in guarddb.items():
            if entry.get('guardstate', False):
                self._driver[channel] = self._driver[channel]
        self._driver.updatePVs()


    def update_request_enum(self, enum, init=False):
        """Update request enum tuple

        Returns True if the enum was changed.

        """
        if enum == guarddb['REQUEST_ENUM']['enums']:
            return False
        guarddb['REQUEST_ENUM']['enums'] = enum
        self._driver.setParamEnums('REQUEST_ENUM', enum)
        if not init:
            # update request enum, in case it's changed
            request = self._driver.getParam('REQUEST')
            if request in guarddb['REQUEST_ENUM']['enums']:
                self._driver.setParam('REQUEST_ENUM', request_enum_index(request))
        self._driver.updatePVs()
        return True



from .system import GuardSystemLoadError

from ._version import __version__

##################################################

def parser_add_system(parser):
    parser.add_argument('system', metavar='<system>', type=str, nargs='?',
                        help="guardian system, by name or module path")
    parser.add_argument('-n', '--name', metavar='<name>', type=str,
                        help="override system name")
    return parser

def parser_add_version(parser):
    parser.add_argument('-v', '--version', action='version',
                        version='Guardian {}'.format(__version__),
                        help="print guardian version and exit")
    return parser

##################################################

def init_system(args, load=False):
    """Initialize (and optionally load) system module

    A SystemExit exception will be thrown if there is a problem
    loading the module.

    """

    if not hasattr(args, 'name'):
        args.name = None
    if not hasattr(args, 'ca_prefix'):
        args.ca_prefix = None

    from . import system

    try:
        sys = system.GuardSystem(args.system,
                                 name=args.name,
                                 ca_prefix=args.ca_prefix)
        if load:
            sys.load()
    except system.GuardSystemError as e:
        raise SystemExit("System error: %s" % e)
    except system.GuardSystemImportError as e:
        raise SystemExit("System import error: %s" % e)
    except GuardSystemLoadError as e:
        raise SystemExit("System load error: %s" % e)

    return sys


def print_states(system, flag_requests=True, requests_only=False, prefix=''):
    """print state enumeration"""
    filters = []
    if requests_only:
        filters.append(('request', True))
    data = list(system.states_iter(filters=filters))
    ilen = max([len(str(s[1])) for s in data])
    for state, index, requestable in data:
        if flag_requests:
            flag = '   '
            if requestable:
                flag = ' * '
        else:
            flag = ' '
        print('{prefix}{index:{ilen}}{flag}{state}'.format(
            index=index,
            state=state,
            flag=flag,
            ilen=ilen,
            prefix=prefix,
        ))


def print_system(system):
    """print system information to stdout"""

    print('ifo: {}'.format(system.ifo))
    print('name: {}'.format(system.name))
    print('module:')
    print('  {}'.format(system.path))

    try:
        system.load()
    except GuardSystemLoadError as e:
        raise SystemExit("System load error: {}".format(e))

    if system.usercode:
        print('usercode:')
        for code in system.usercode:
            print('  {}'.format(code))
    print('CA prefix: {}'.format((system.ca_prefix or '')))
    if system.ca_monitor:
        print('CA monitor: {}'.format(system.ca_monitor))
        print('CA monitor notify: {}'.format(system.ca_monitor_notify))
    print('nominal state: {}'.format((system.nominal or 'NONE')))
    print('initial request: {}'.format((system.request or 'NONE')))
    print('states (*=requestable):')
    print_states(system, prefix='  ')



import re
import inspect
import networkx
from networkx.drawing.nx_pydot import to_pydot

NOT_DECORATORS = ['property']

def _get_jump_edges(obj):
    dest_nodes = set()
    for source_line in inspect.getsourcelines(obj)[0]:
        match = re.match(r"\s*return '(.*)'\s*$", source_line)
        if match:
            dest_nodes.update([match.group(1)])
    return dest_nodes

def _is_comment_or_blank(source_line):
    return re.match(r"\s*#", source_line) or re.match(r"^\s*$", source_line)

def find_jump_edges(guardstate, module):
    """Find all jump edges for a GuardState object."""

    dest_nodes = _get_jump_edges(guardstate)

    source_lines = inspect.getsourcelines(guardstate)[0]
    i = 0
    while i < len(source_lines):
        # look for jump edges in object decorator source code
        # decorators can be stacked on top of each other, so we need to
        # be sure to get all of them
        decorator_matches = set()
        j = i
        while True:
            match = re.match(r"\s*@([^\s\(]*)(\(.*\))?\s*$", source_lines[j])
            if _is_comment_or_blank(source_lines[j]):
                pass
            elif match:
                decorator_matches.update([match.group(1)])
            else:
                break
            j += 1

        # we know these are decorators if they are followed by a function definition
        if decorator_matches:
            while True:
                if _is_comment_or_blank(source_lines[j]):
                    pass
                elif re.match(r".*\bdef\b.*", source_lines[j]):
                    # look for jump edges in decorator code
                    for decorator_match in decorator_matches:
                        if decorator_match in NOT_DECORATORS:
                            continue
                        dest_nodes.update(_get_jump_edges(getattr(module, decorator_match)))
                        break
                else:
                    break
                j += 1

        i = j+1

    return dest_nodes

def fullgraph(system, find_jumps=False):
    """Return MultiDiGraph copy of system graph with JUMP edges added."""
    G = networkx.MultiDiGraph(system.graph)
    for node, code in G.nodes(data='code'):
        obj = code()
        if find_jumps:
            for jump in find_jump_edges(code, system.module):
                G.add_edge(node, jump, jump=True, goto=False, weight=1)
    return G

def sys2dot(system,
            path=None,
            show_index=False,
            show_gotos=False,
            show_jumps=False,
            edge_constraints=True):
    """Create graphviz dot code for system graph."""

    # if states is a tuple it specifies a start and stop state, so
    # calc path
    fpath = []
    path_edges = []
    if path:
        assert isinstance(path, tuple), "'path' should be tuple"
        if len(path) > 1:
            fpath = system.shortest_path(path[0], path[1])
        else:
            fpath = path

        for ind in range(len(fpath)-1):
            path_edges.append((fpath[ind], fpath[ind+1]))

    G = fullgraph(system, find_jumps=show_jumps)

    G.graph['graph'] = {
        'label': '<%s Guardian: %s>' % (system.ifo, system.name),
        'labelloc': 't',
        'rankdir': 'BT',
    }

    # http://www.graphviz.org/content/attrs
    for node, data in G.nodes(data=True):
        # node attribute defaults
        label = '%s' % (node)
        shape = 'oval'
        style = 'filled'
        colorscheme = 'set19'
        color = '2'
        outline = 'black'
        fontcolor = 'white'
        penwidth = 1

        # state index
        if show_index:
            label = '%s (%d)' % (node, data['index'])

        # if not requestable use pastel shade
        if not data['request']:
            colorscheme = 'pastel19'
            fontcolor = 'grey30'

        # if not redirectable use red outline
        if not data['redirect']:
            outline = 'red'
            #style = 'dashed, filled'
            penwidth = 2

        # if node is goto use green
        if data['goto']:
            color = '3'

        # INIT state
        if node == 'INIT':
            shape = 'doublecircle'
            color = '4'

        # if node in path
        if fpath and node in fpath:
            penwidth = 8
            colorscheme = 'set19'
            if node == fpath[0]:
                color = 3
            elif node == fpath[-1]:
                color = 1
            else:
                color = 5

        # set attributes
        G.nodes[node]['label'] = label
        G.nodes[node]['shape'] = shape
        G.nodes[node]['style'] = style
        G.nodes[node]['fillcolor'] = '/%s/%s' % (colorscheme, color)
        G.nodes[node]['penwidth'] = penwidth
        G.nodes[node]['color'] = outline
        G.nodes[node]['fontcolor'] = fontcolor

    # set edge attributes
    remove_edges = []
    for edge in G.edges(data=True, keys=True):
        source, sink, key, data = edge

        # label edges with their weight
        if data['weight'] != 1:
            data['label'] = data['weight']

        # graphviz handles edge weights opposite of networkx.  edge
        # weights add to path length in networkx whereas large weights
        # represent "tighter" couplings between nodes in graphviz.
        # therefore flip the sign of the weight for graphviz, so that
        # larger weights represent weaker coupling
        data['weight'] *= -1

        # goto edges
        if data['goto']:
            if show_gotos:
                if not edge_constraints:
                    data['constraint'] = False
                data['color'] = 'grey'
                data['fontcolor'] = 'grey'
            else:
                if edge[:2] not in path_edges:
                    remove_edges.append((source, sink, key))

        # jump edges
        if data.get('jump', False):
            if not edge_constraints:
                data['constraint'] = False
            data['style'] = 'dashed'
            data['color'] = 'red'

        # path edges
        if edge[:2] in path_edges:
            data['color'] = '/set19/5'
            data['penwidth'] = 5

    for edge in remove_edges:
        G.remove_edge(*edge)

    # do the graphviz conversion
    dot = to_pydot(G)

    return dot



import threading
import functools

##################################################

class GuardState(object):
    """Guardian system state machine state.

    Execution model:

    This type of state is known as a "Moore" state: the primary action
    is executed when the state is entered (main() method).

    The main() method is executed once upon entering the state, after
    which the run() method is executed in a loop.  The return values
    of both methods are interpreted the same, as follows:

    If the return value of either method is True, the state will
    "complete" (status: DONE).  If the state is not the requested
    state the system will immediately transition to the target state.
    If the state is equal to the requested state the run method will
    be executed.

    If the return value is a string it will be interpreted as a state
    name and the system will transition immediately to the new state.
    This is known as a "jump" transition.

    If the return value is None or False, the run() method is
    executed.  NOTE: if unspecified, functions return None by default.

    State properties (user specified):

    request: indicates whether or not this state is requestable by the
    user.  Default: True

    goto: if True will cause edges to be added to this state from all
    states in the system graph.  Default: False

    index: numeric index for state.  must be a positive definite
    number.  If not specified, a negative index will be automatically
    assigned.

    State objects:

    timer: TimerManager object, used to count down a specified amount
    of time.

    >>> self.timer['foo'] = 3
    >>> self.timer['foo']
    False
    >>> time.sleep(3)
    >>> self.timer['foo']
    True

    The environment of state method execution includes a couple of
    additional "built-in" functions:

    log: write to state the system logger:

    >>> log("doing something")

    ezca: Ezca EPICS channel access:

    >>> ezca['A'] = 3

    """
    def __init__(self, logfunc=None):
        self._logfunc = logfunc
        self.timer = TimerManager(logfunc=logfunc)

    @property
    def name(self):
        return self.__class__.__name__

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.name)

    def __str__(self):
        # use base name
        return "<GuardState %r>" % (self.name)

    ##########

    def log(self, message):
        if self._logfunc:
            self._logfunc(message)
        else:
            print(message)

    ##########
    ##########
    # user overridable functions/variables

    request = True
    """Is this state "requestable"."""

    goto = False
    """Is this a "goto" state.

    If specified as :type int, the value will be used as the edge
    weight for all goto edges to this state.
    """

    redirect = True
    """Can this state be "redirected".

    If False the state is "protected" and redirects away from this
    state are ignored while the state is returning False.

    """

    ##########
    def main(self):
        """state MAIN method.

        This method is executed once immediately upon entering state.

        See "Execution model" above for more info.

        """
        return False

    ##########
    def run(self):
        """state RUN method.

        This method is executed in a loop if the main() method returns
        False or None.

        See "Execution model" above for more info.

        """
        return True

    ##########
    ##########

    def destroy(self):
        self.timer.destroy()
        # FIXME: anything else?

##################################################

class TimerError(Exception):
    pass

class Timer(object):
    """Simple single timer.

    Once initialized with a time in seconds, the done property returns
    False until the timeout, after which done returns True.

    """
    def __init__(self, time, name=None, logfunc=None):
        self._name = name
        self._logfunc = logfunc
        self._done = False
        self._timer = threading.Timer(time, self._stop)
        self._timer.daemon = True
        self._timer.start()

    def _stop(self):
        self._done = True
        if self._logfunc and self._name:
            self._logfunc("timer['%s'] done" % self._name)

    @property
    def done(self):
        """Return True if timeout reached, False otherwise."""
        return self._done

    def cancel(self):
        """Cancel timer."""
        self._timer.cancel()

class TimerManager(object):
    """Manage named timers.

    Setitem initializes and starts a named timer.
    Getitem returns timeout status.

    """
    def __init__(self, logfunc=None):
        self._logfunc = logfunc
        self._timers = {}

    def __setitem__(self, name, time):
        """Set named timer for seconds."""
        if name in self._timers:
            self._timers[name].cancel()
        self._timers[name] = Timer(time, name=name, logfunc=self._logfunc)
        if self._logfunc:
            self._logfunc("timer['%s'] = %s" % (name, time))

    def __getitem__(self, name):
        """Return True if named timer done, False otherwise."""
        if name not in self._timers:
            raise TimerError("timer '%s' not initialized." % name)
        return self._timers[name].done

    def destroy(self):
        """Destroy all timer objects."""
        for timer in self._timers.values():
            timer.cancel()

##################################################

class GuardStateDecorator(object):
    """GuardState method decorator base class.

    GuardStateDecorator is a function decorator designed for
    GuardState methods.  It executes "check" code before and/or after
    the function being wrapped.  It has three user-overrideable
    methods, *pre_exec*, *pre_return_exec*, and *post_exec*.  If any
    of these methods return a value that is not None, those return
    values will be returned from the wrapped function call.  See the
    individual method help for more information.

    Example: The following would cause the FOO.run() method to return
    'TRIPPED' immediately if the 'WD' channel equals 0.  The rest of
    the FOO.run() method will not be executed:

    class check_wd(GuardStateDecorator):
        def pre_exec(self):
            if ezca['WD'] == 0:
                return 'TRIPPED'

    class FOO(GuardState):
        @check_wd
        def run(self):
            ...

    """
    def __init__(self, func):
        functools.wraps(func)(self)
        self.func = func

    def __get__(self, state_obj, owner_class=None):
        return functools.partial(self.__call__, state_obj)

    def __call__(self, state_obj, *args, **kwargs):
        pre_exec_return = self.pre_exec()
        if pre_exec_return is not None:
            return pre_exec_return

        main_return = self.func.__call__(state_obj, *args, **kwargs)

        pre_return_exec_return = self.pre_return_exec()
        if pre_return_exec_return is not None:
            return pre_return_exec_return

        if main_return is not None:
            return main_return

        post_exec_return = self.post_exec()
        if post_exec_return is not None:
            return post_exec_return

        return main_return

    # FIXME: should these all be staticmethods?

    def pre_exec(self):
        """Pre-execution function.

        Executed before the wrapped function.  If this function's
        return value is not None, its return value will be returned
        immediately and none of the other functions will be executed.

        """
        pass

    def pre_return_exec(self):
        """Pre-return function.

        Executed after the wrapped function, but before inspection of
        the wrapped function's return value.  If this function's
        return value is not None, its return value will be returned
        immediately, the wrapped function's return value will be
        ignored, and *post_exec* will not be executed.

        """
        pass

    def post_exec(self):
        """Post-execution function.

        Executed after the wrapped function.  If none of the previous
        methods, or the wrapped function, have return values that are
        not None, this method will be executed, and its return value
        will be used if it is not None.

        """
        pass



from __future__ import print_function
import os
import sys
import shutil
import argparse
import tempfile
import traceback
import subprocess
from termcolor import colored

from . import cli
from . import const
from .ligopath import ARCHIVE_ROOT

############################################################

def nds_connection():
    import nds2
    port = 31200
    NDSSERVER = os.getenv('NDSSERVER')
    if NDSSERVER:
        hostport = NDSSERVER.split(',')[0].split(':')
        host = hostport[0]
        if len(hostport) > 1:
            port = int(hostport[1])
    else:
        raise ValueError('NDSSERVER not specified')
    return nds2.connection(host, port)

############################################################

parser = argparse.ArgumentParser(
    prog = 'guardutil',
    description = "Advanced LIGO Guardian system utility.",
    formatter_class = argparse.RawDescriptionHelpFormatter,
    epilog = """

Add '-h' after individual commands for command help.

Environment variables:
  IFO                   IFO designator (required for most operations)
  SITE                  SITE designator
  USERAPPS_DIR          LIGO "userapps" root path [/opt/rtcds/userapps/release]
  GUARD_MODULE_PATH     override default userapps system module search path
  GUARD_ARCHIVE_ROOT    system archive root directory (may be URL)
""",
    )

cli.parser_add_version(parser)

subparser = parser.add_subparsers(
    title='Commands',
    metavar='<command>',
    dest='cmd',
    #help=argparse.SUPPRESS,
    )
subparser.required = True

def gen_subparser(cmd, func, sys=True):
    help = func.__doc__.split('\n')[0].lower().strip('.')
    desc = func.__doc__
    p = subparser.add_parser(cmd,
                             formatter_class=argparse.RawDescriptionHelpFormatter,
                             help=help,
                             description=desc)
    if sys:
        cli.parser_add_system(p)
    p.set_defaults(func=func)
    return p

############################################################

def command_help(args):
    """Show command help."""
    if args.cmd:
        try:
            p = subparser.choices[args.cmd]
        except KeyError:
            sys.exit("Unknown command: {}".format(args.cmd))
    else:
        p = parser
    p.print_help()

p = gen_subparser("help", command_help, sys=False)
p.add_argument('cmd', metavar='<command>', nargs='?', help="command")


def print_system(args):
    """Print general system information."""
    system = cli.init_system(args, load=False)
    cli.print_system(system)

p = gen_subparser("print", print_system)


def print_states(args):
    """Print state enumeration."""
    system = cli.init_system(args, load=True)
    cli.print_states(system, flag_requests=False,
                     requests_only=args.requests_only)

p = gen_subparser("states", print_states)
p.add_argument('-r', '--requests-only', action='store_true',
               help="output request states only")


def print_edges(args):
    """Print transition edges."""
    system = cli.init_system(args, load=True)
    for edge in system.graph.edges(data=True):
        source, sink, data = edge
        print('({0}, {1}, {2})'.format(source, sink, data['weight']))

p = gen_subparser("edges", print_edges)


def print_path(args):
    """Print system graph path as a list of states."""
    system = cli.init_system(args, load=True)
    for state in system.shortest_path(args.state0, args.state1):
        print(state)

p = gen_subparser("path", print_path)
p.add_argument('state0', help="initial state")
p.add_argument('state1', help="final state")


def draw_graph(args):
    """Draw/display system graph."""
    try:
        import guardian.graph as graph
    except ImportError as e:
        sys.exit("Graph drawing not supported: %s" % e)
    system = cli.init_system(args, load=True)
    if args.all:
        args.gotos = True
        args.jumps = True
    path = ()
    if args.state:
        if args.request:
            path = (args.state, args.request)
        else:
            path = (args.state,)
    for state in path:
        if state not in system:
            sys.exit("State '%s' not defined in system %s." % (state, system.name))

    dot = graph.sys2dot(system,
                        path=path,
                        show_index=args.index,
                        show_gotos=args.gotos,
                        show_jumps=args.jumps,
                        edge_constraints=args.constraints,
                        )

    # display graph or write to file
    default_format = 'pdf'
    if args.outfile is not False:
        outfile = '%s_%s' % (system.ifo, system.name)
        outpath = args.outfile
        if outpath is None:
            outpath = outfile
        elif os.path.isdir(os.path.expanduser(outpath)):
            outpath = os.path.join(os.path.expanduser(outpath), outfile)
        ext = os.path.splitext(outpath)[1]
        if args.format:
            fmt = args.format
        elif ext != '':
            fmt = ext.strip('.')
        else:
            fmt = default_format
        if ext == '':
            outpath += '.'+fmt
        dot.write(outpath, format=fmt)
    else:
        if not args.format:
            args.format = default_format
        suffix = '.%s' % args.format
        if args.format == 'pdf':
            viewer = ['evince', '-w']
        elif args.format == 'svg':
            viewer = ['inkview']
            #viewer = ['iceweasel']
        elif args.format == 'dot':
            viewer = ['xdot']
        else:
            viewer = ['xdg-open']
        try:
            # FIXME: better temp file name
            with tempfile.NamedTemporaryFile(suffix=suffix, delete=False) as f:
                dot.write(f.name, format=args.format)
                try:
                    subprocess.run(viewer + [f.name])
                except OSError as e:
                    sys.exit("Graph viewer '%s' could not be found." % (viewer[0]))
        except KeyboardInterrupt:
            sys.exit()

p = gen_subparser("graph", draw_graph)
p.add_argument('-i', '--index', action='store_true',
               help="add state indices to labels")
p.add_argument('-g', '--gotos', action='store_true',
               help="draw 'goto' edges")
p.add_argument('-j', '--jumps', action='store_true',
               help="draw 'jump' edges")
p.add_argument('-a', '--all', action='store_true',
               help="show all edges (-gj)")
p.add_argument('-c', '--constraints', action='store_true',
               help="use edge constraints for goto and jump edges")
p.add_argument('-f', '--format', metavar='<type>', type=str,
               help="drawing format: 'pdf', 'svg', etc. (default: 'pdf')")
p.add_argument('-o', '--outfile', metavar='<path>', type=str, nargs='?', default=False,
               help="save graph to file.  If path is a directory, a file named \"<system name>.<format>\" will be saved in that directory.  If path includes extension the extension will be used to determine format.")
p.add_argument('state', metavar='<state>', type=str, nargs='?',
               help="state to highlight")
p.add_argument('request', metavar='<request>', type=str, nargs='?',
               help="final (request) state for highlighting path")


def print_files(args):
    """Print paths to all usercode source files."""
    system = cli.init_system(args, load=True)
    print(system.path)
    for path in system.usercode:
        print(path)

p = gen_subparser("files", print_files)


def print_source(args):
    """Print usercode source lines for system object."""
    import inspect
    system = cli.init_system(args, load=True)
    try:
        obj = system.module.__dict__[args.object]
    except KeyError as e:
        sys.exit("Unknown system object: %s" % e)
    sf = inspect.getsourcefile(obj)
    lines, sl = inspect.getsourcelines(obj)
    el = sl + len(lines) - 1
    nwidth = len(str(sl + len(lines)))
    if not args.raw:
        print("object: %s" % (obj))
        print("  file: %s" % (sf))
        print(" lines: %d-%d" % (sl, el))
        print('-' * (nwidth + len(lines[0])))
    prefix = ''
    for line in lines:
        if not args.raw:
            prefix = \
                colored('{sl:{nwidth}}'.format(sl=sl, nwidth=nwidth), 'green') + \
                colored(':', 'cyan')
            sl += 1
        print('%s%s' % (prefix, line), end='')

p = gen_subparser("source", print_source)
p.add_argument('object', metavar='<object>',
               help="code object")
p.add_argument('-r', '--raw', action='store_true',
               help="raw source, without header or line numbers")


def edit_system(args):
    """Edit system usercode."""
    try:
        system = cli.init_system(args, load=True)
    except:
        print("WARNING: Exception encountered when loading module:\n", file=sys.stderr)
        print(traceback.format_exc(), file=sys.stderr)
        system = cli.init_system(args, load=False)
    cmd = []
    if args.editor:
        cmd = args.editor.split()
    elif os.getenv('EDITOR'):
        cmd = os.getenv('EDITOR').split()
    else:
        # try to find XDG-specified handler for text/x-python
        from gi.repository import Gio
        mimetype = 'text/x-python'
        # XDG DATA directories
        # http://standards.freedesktop.org/basedir-spec/basedir-spec-latest.html
        xdg_share = [os.getenv('XDG_DATA_HOME', os.path.expanduser('~/.local/share'))] \
                    + os.getenv('XDG_DATA_DIRS', '/usr/local/share:/usr/share').split(':')
        # get desktop handler for mimetype
        desktop = subprocess.check_output(
            ['xdg-mime', 'query', 'default', mimetype],
            universal_newlines=True,
        ).strip()
        # look for desktop file
        if desktop:
            for d in xdg_share:
                path = os.path.join(d, 'applications', desktop)
                if not os.path.exists(path):
                    continue
                # extract executable
                app = Gio.DesktopAppInfo.new_from_filename(path)
                exe = app.get_executable()
                if exe:
                    cmd = [exe]
                    break
    if not cmd:
        cmd = ['emacs']
    if args.object:
        import inspect
        try:
            obj = system.module.__dict__[args.object]
        except KeyError as e:
            sys.exit("Unknown system object: %s" % e)
        source_file = inspect.getsourcefile(obj)
        source_lines, line_start = inspect.getsourcelines(obj)
        line_end = line_start + len(source_lines)
        #cmd += ['+%d' % sl, sf]
        if 'emacs' in cmd[0]:
            cmd += [source_file,
                    '--eval',
                    "(narrow-to-region (line-beginning-position {}) (line-end-position {}))".format(line_start, line_end),
                    ]
        elif 'gedit' in cmd[0]:
            cmd += [source_file, '+{}'.format(line_start)]
        else:
            sys.exit("Object editing only supported with emacs and gedit.")
    else:
        cmd += [system.path]
        if system.usercode:
            cmd += system.usercode
    print(' '.join(cmd), file=sys.stderr)
    try:
        try:
            subprocess.run(cmd)
        except OSError:
            sys.exit("Could not find editor: %s" % cmd[0])
    except KeyboardInterrupt:
        sys.exit()

p = gen_subparser("edit", edit_system)
p.add_argument('object', metavar='<object>', type=str, nargs='?',
               help="object code to jump to in editor")
p.add_argument('-e', '--editor', type=str,
               help="editor (emacs, gedit, etc. (default: xdg-open))")


def archive_clone(args):
    """Clone system usercode git archive for inspection.

The GUARD_ARCHIVE_ROOT environment variable points to the root
directory where all system archives are stored.  This may point to at
network location as well (e.g. https://...)."""
    if not ARCHIVE_ROOT:
        sys.exit("GUARD_ARCHIVE_ROOT or IFO/SITE env vars must be specified to find code archives.")

    import git

    archive_path = os.path.join(ARCHIVE_ROOT, args.system)

    if '://' not in archive_path:
        archive_path = os.path.abspath(archive_path)

    tempdir = None
    if args.directory:
        if not os.path.exists(args.directory):
            os.mkdir(args.directory)
        git.Git(args.directory).clone(archive_path)
    else:
        tempdir = tempfile.mkdtemp(prefix='guardutil_archive_%s_' % (args.system))
        try:
            wd = os.path.join(tempdir, args.system)
            os.mkdir(wd)
            #git.Repo.clone_from(archive_path, root)
            git.Git(tempdir).clone(archive_path)
            os.chdir(wd)
            if args.shell:
                subprocess.run('bash')
            else:
                subprocess.run('gitk')
        except KeyboardInterrupt:
            pass
        finally:
            if tempdir:
                shutil.rmtree(tempdir)

p = gen_subparser("archive-clone", archive_clone)
# requires name only (not path) so create separately
p.add_argument('directory', metavar='<directory>', nargs='?',
               help="checkout directory")
p.add_argument('-s', '--shell', action='store_true',
               help="open shell in checked-out repo")


def print_history(args):
    """Print state history around specified time (NDS)."""
    from gpstime import gpstime
    system = cli.init_system(args, load=True)
    channel = const.CAS_PREFIX_FMT.format(IFO=system.ifo, SYSTEM=system.name) + 'STATE_N'
    time0 = gpstime.parse(args.time0)
    if args.time1:
        t0 = int(time0.gps())
        t1 = int(gpstime.parse(args.time1).gps())
    else:
        window = eval(args.window)
        t0 = int(time0.gps() + window[0])
        t1 = int(time0.gps() + window[1])
    print("NDS fetch: {} -> {}".format(t0, t1), file=sys.stderr)
    conn = nds_connection()
    buf = conn.fetch(t0, t1, [channel])
    buf = buf[0]
    state_array = buf.data

    # some helper function
    def i2t(i):
        """gps time of NDS data index"""
        return buf.gps_seconds + i*(1./buf.channel.sample_rate)
    def stime(gt):
        """gpstime string"""
        if args.tfmt == 'gps':
            return '{:.6f}'.format(gt.gps())
        elif args.tfmt == 'delta':
            td = gt.gps() - time0.gps()
            return '{:.3f}'.format(td)
        else:
            return gt.iso()
    def ptrans(last, cur):
        """print transition"""
        print('{time} {frm} -> {to}'.format(
            time=stime(cur[1]), frm=last[0], to=cur[0]))
    def pdur(last, cur):
        """print duration"""
        print('{state:{msl}s} {ent} -> {ext} ({dur})'.format(
            msl=msl, state=last[0],
            ent=stime(last[1]), ext=stime(cur[1]), dur=(cur[1] - last[1])))

    # find all transitions
    trans = []
    last = None
    for i, s in enumerate(state_array):
        if s == last:
            continue
        trans.append((
            system.index(int(s)),
            gpstime.fromgps(i2t(i)),
            ))
        last = s
    # record the last sample
    trans.append((
        system.index(int(s)),
        gpstime.fromgps(i2t(i)),
        ))
    msl = max([len(s) for s,t in trans])
    # print the results
    last = trans[0]
    for cur in trans[:-1]:
        if cur == last:
            continue
        if args.transitions:
            ptrans(last, cur)
        else:
            pdur(last, cur)
        last = cur
    if not args.transitions:
        pdur(last, trans[-1])

p = gen_subparser("print-hist", print_history)
p.add_argument('-t', '--transitions', action='store_true',
               help="show state transitions rather than duration")
p.add_argument('-g', '--gps', action='store_const', dest='tfmt', const='delta',
               help="show times in GPS")
p.add_argument('-d', '--delta', action='store_const', dest='tfmt', const='delta',
               help="show times as delta relative to time0")
p.add_argument('-w', '--window', metavar='<window>', default='[-60, 60]',
               help="time window if time1 not specified (default: [-60, 60])")
p.add_argument('time0', metavar='<time0>',
               help="center time, or start time if time1 specified")
p.add_argument('time1', metavar='<time1>', nargs='?',
               help="stop time")


def plot_history(args):
    """Plot node data/state around specified time (NDS)."""
    import guardian.plotutils as pu
    system = cli.init_system(args, load=True)
    center_time = args.time0
    window = eval(args.window)
    pu.guardfig(system, center_time, window)

p = gen_subparser("plot-hist", plot_history)
p.add_argument('-w', '--window', metavar='<window>', default='[-60, 60]',
               help="time window if time1 not specified (default: [-60, 60])")
p.add_argument('time0', metavar='<time0>',
               help="center or start time")
p.add_argument('time1', metavar='<time1>', nargs='?',
               help="duration or stop time")

############################################################

def main():
    args = parser.parse_args()
    # try:
    #     system = cli.init_system(args, load=True)
    # except:
    #     print("WARNING: Exception encountered when loading module:\n", file=sys.stderr)
    #     print(traceback.format_exc(), file=sys.stderr)
    #     system = cli.init_system(args, load=False)
    # arg_filter = ['cmd', 'func', 'system', 'name']
    # cargs = {k:v for k,v in args.__dict__.items() if k not in arg_filter}
    # print(cargs)
    args.func(args)

if __name__ == '__main__':
    main()



import os
import sys
import signal
import argparse
if sys.version_info.major > 2:
    import builtins
else:
    import __builtin__ as builtins

from . import db
from . import cli
from . import const
from .log import get_logger

############################################################

PROG = 'guardian'
description = '''Advanced LIGO Guardian state machine engine.'''
usage = '''
guardian [<options>] [-i <system>]                 interactive shell
guardian [<options>] <system>                      state machine daemon
guardian [<options>] <system> <state>              single state execution
guardian [<options>] <system> <state> <request>    single path execution
'''
fulldesc = """
Guardian is a state machine traversal engine.  State machines are
described by system description modules.  See the following link for a
full description of Guardian, including module syntax, and state
machine behavior:

  https://awiki.ligo-wa.caltech.edu/aLIGO/Guardian

If no system is specified or the interactive flag is provided the
interactive interpreter will be launched.  This is useful for testing
commands, including interactive use of the Ezca EPICS interface.

Systems can be specified either by full path to the system description
module, or by name (<NAME>).  If provided as path, the path basename
(file name minus '.py' extension) will be used as the system name.  If
specified by name, guardian will search for the appropriate module in
GUARD_MODULE_PATH, which defaults to:

  <USERAPPS_DIR>/*/{<ifo>,common}/guardian.

When a single system argument is provided the state machine engine
daemon is launched, which immediately begins executing the state graph
described in the module.  Control and status of the daemon is provided
via EPICS channels with the prefix '<IFO>:GRD_<NAME>_'.

Additional arguments are interpreted as system state names.  If a
single state name is supplied, the single state will be executed to
completion, after which the program will exit.  If two states are
specified the first will be interpreted as initial state and the
second as requested state.  Guardian will attempt to reach the
requested state by executing the shortest path through the graph.
Upon completion of the requested state the program will exit.

Ctrl-C can be used to stop execution at any point.

The --name and --ca-prefix options can be used to override the system
name and prefix specifications.  Prefixes are combined with the IFO
environment variable to create a proper LIGO channel name prefix that
will be appended to all EPICS channel access calls via the ezca
object (see Ezca for more info).

Environment variables:
  IFO                       IFO designator (required)
  USERAPPS_DIR              LIGO USERAPPS root path [/opt/rtcds/userapps/release]
  GUARD_MODULE_PATH         Override default userapps system module search path
  GUARD_ARCHIVE_ROOT        Set system archive root directory, enable code archiving
  GUARD_LOG_LEVEL           Log level: WARNING, [INFO], DEBUG, LOOP
  GUARD_CPS                 Main loop cycles per seconds [16]
  EZCA_TIMEOUT              Channel access connection timeout in seconds [2]
  EZCA_CA                   Channel access enable/disable [TRUE]
  EPICS_CAS_INTF_ADDR_LIST  Network interface for GRD interface
  EPICS_CAS_SERVER_PORT     Network port for GRD interface"""

parser = argparse.ArgumentParser(
    formatter_class=argparse.RawDescriptionHelpFormatter,
    prog=PROG,
    usage=usage,
    )

cli.parser_add_system(parser)
parser.add_argument('state', metavar='<state>', type=str, nargs='?',
                    help="initial state")
parser.add_argument('request', metavar='<request>', type=str, nargs='?',
                    help="request state")

parser.add_argument('-ca', '--prefix', '--ca-prefix', metavar='<prefix>',
                    type=str, dest='ca_prefix',
                    help="system EPICS channel access prefix ")
parser.add_argument('-s', '--state', metavar='<state>', dest='dstate',
                    help="set initial state for daemon")
parser.add_argument('-r', '--request', metavar='<state>', dest='drequest',
                    help="set initial request state for daemon")
parser.add_argument('-o', '--op', metavar='<op>', dest='dop',
                    choices=db.DAEMON_OP, default='EXEC',
                    help="set initial daemon OP ({}|{})".format(*db.DAEMON_OP))
parser.add_argument('-m', '--mode', metavar='<mode>', dest='dmode',
                    choices=db.DAEMON_MODE,
                    help="set initial daemon MODE ({}|{}|{})".format(*db.DAEMON_MODE))

cgroup = parser.add_mutually_exclusive_group()
cgroup.add_argument('-i', '--interactive', action='store_true',
                    help="launch interactive guardian shell")
cgroup.add_argument('-p', '--print', action='store_const', dest='cmd', const='print',
                    help="print system information and exit")
cgroup.add_argument('-c', '--channels', action='store_const', dest='cmd', const='channels',
                    help="print daemon CAS channels (archive channels marked with '*')")

cli.parser_add_version(parser)
parser.add_argument('--usage', action='store_true',
                    help="show more detailed usage and exit")

############################################################

def main():

    args = parser.parse_args()

    if args.usage:
        print(description)
        print(fulldesc)
        sys.exit()

    ##############################################

    system = None
    if args.system:
        system = cli.init_system(args)

    ##############################################

    if args.cmd == 'channels':
        ifo = '<IFO>'
        name = '<NODE>'
        if system is not None:
            if system.ifo:
                ifo = system.ifo
            name = system.name
        for chan in sorted(db.guarddb):
            channel = const.CAS_PREFIX_FMT.format(IFO=ifo, SYSTEM=name) + chan
            flag = ''
            if db.guarddb[chan].get('archive'):
                flag = '*'
            print('{} {}'.format(channel, flag))
        sys.exit()

    if args.cmd == 'print':
        if system is None:
            sys.exit("Must specify system to print.")
        cli.print_system(system)
        sys.exit()
    elif system is not None:
        system.load()

    if args.interactive or not args.system:
        import IPython

        from .state import GuardState, GuardStateDecorator
        from .manager import Node, NodeManager

        from ezca import Ezca, LIGOFilter, LIGOFilterManager

        banner = """
--------------------
aLIGO Guardian Shell
--------------------
""".lstrip()

        builtins._IFO_ = const.IFO

        user_ns = {
            'IFO': _IFO_,
            '_IFO_': _IFO_,
            'LIGOFilter': LIGOFilter,
            'LIGOFilterManager': LIGOFilterManager,
            'GuardState': GuardState,
            'GuardStateDecorator': GuardStateDecorator,
            'Node': Node,
            'NodeManager': NodeManager,
        }

        from .ligopath import userapps_guardian_paths
        sys.path += userapps_guardian_paths()

        if system:
            builtins._SYSTEM_ = system.name
            builtins.ezca = Ezca(prefix=system.ca_prefix)
        else:
            builtins._SYSTEM_ = const.HOSTNAME
            builtins.ezca = Ezca(prefix=args.ca_prefix)

        user_ns['SYSTEM'] = _SYSTEM_
        user_ns['_SYSTEM_'] = _SYSTEM_

        user_ns['ezca'] = ezca
        if ezca.prefix:
            banner += "ezca prefix: %s\n" % ezca.prefix

        if system:
            banner += "system: %s (%s)\n" % (system.name, system.path)
            user_ns[system.name] = system
            user_module = system.module
            user_module.__dict__.update(user_ns)
            user_ns = None
            if system.node_managers:
                banner += "node managers ('<name>.init()' to initialize):\n"
                for name, obj in system._node_managers.items():
                    banner += '  {} = {!r}\n'.format(name, obj)
        else:
            user_module = None

        def notify(msg):
            print('USERMSG: {}'.format(msg))
        def log(msg):
            print(msg)

        builtins.notify = notify
        builtins.log = log

        IPython.embed(banner1=banner, user_module=user_module, user_ns=user_ns)
        sys.exit()

    ##############################################
    ##############################################

    try:
        from .daemon import Daemon
    except ImportError as e:
        sys.exit("Guardian daemon not supported on this system: %s" % e)

    single_shot = False
    loglevel = 'INFO'

    if args.state:
        single_shot = True

        if not args.request:
            args.request = args.state

    # validate states if specified
    initial_state = args.state or args.dstate
    initial_request = args.request or args.drequest
    for state in initial_state, initial_request:
        if state and state not in system:
            sys.exit("State '%s' not defined in system %s." % (state, system.name))

    # specify archive if not running single shot
    if os.getenv('GUARD_ARCHIVE_ROOT', None) and not single_shot:
        from .archive import SystemArchiveGit
        archive_dir = os.path.join(os.getenv('GUARD_ARCHIVE_ROOT'), system.name)
        archive = SystemArchiveGit(archive_dir)
    else:
        archive = None

    ##############################################

    logger = get_logger(system.name, level=loglevel)

    guard = Daemon(system, logger,
                   initial_op=args.dop,
                   initial_mode=args.dmode,
                   initial_state=args.state or args.dstate,
                   initial_request=args.request or args.drequest,
                   single_shot=single_shot,
                   archive=archive,
                   )

    # set TERM signal handler
    # FIXME: this shouldn't be necessary since daemon=True should
    # take care of this
    def sigterm(signum, frame):
        sys.stderr.write("caught signal %s; terminating\n" % signum)
        raise SystemExit
    signal.signal(signal.SIGTERM, sigterm)

    try:
        guard.run()
    except KeyboardInterrupt:
        sys.stderr.write("caught keyboard interrupt; terminating\n")
    finally:
        guard.stop()



if __name__ == '__main__':
    main()



import os
import re
import fnmatch
import logging
import subprocess
from gpstime import gpstime
from dateutil.tz import tzlocal

from . import util

##################################################

NAME = os.getenv('GUARDNAME', 'guardian')
SYSTEMCTL_CMD = ['systemctl', '--user']
GUARD_TARGET_PATH = os.path.expanduser('~/.config/systemd/user/guardian.target.wants')
UNIT_FMT = '{}@{{}}.service'.format(NAME)
UNIT2NODE_RE = re.compile(UNIT_FMT.format('(.*)'))
JOURNALCTL_TIME_FMT = '%Y-%m-%d %X'

##################################################

class SystemctlError(Exception):
    pass

def systemctl_cmd(*args, nodes='*', check=True, **kwargs):
    cmd = SYSTEMCTL_CMD + list(args)
    if nodes:
        cmd += [UNIT_FMT.format(node) for node in nodes]
    logging.debug('cmd: ' + ' '.join(cmd))
    try:
        proc = subprocess.run(cmd, check=check, **kwargs)
    except subprocess.CalledProcessError as e:
        raise SystemctlError(e)
    return proc

##################################################

def list_nodes(node_filter='*', only_active=False, include_status=False):
    """List nodes and node status.

    A list of node name glob patterns can be supplied in `node_filter`
    to retrict the return nodes to only those matching filters.  By
    default all nodes are returned (filter='*').  If node_filter is
    None or an empty list, then no nodes will be returned.

    If `only_active` is True, only active nodes will be returned.
    If `include_status` is True, list elements will be tuples of node
    (name, status).

    """
    # if not nodes:
    #     nodes = '*'
    # proc = systemctl_cmd('list-units', '--full', '--no-legend', nodes=nodes,
    #                      stdout=subprocess.PIPE, universal_newlines=True, bufsize=1,
    #                      )
    # for line in proc.stdout.split('\n'):
    #     if not line:
    #         continue
    #     unit, load, active, sub, desc = line.split(maxsplit=4)
    #     node_match = UNIT2NODE_RE.match(unit)
    #     # FIXME: should we skip any weird output?
    #     if not node_match:
    #         continue
    #     node = node_match.group(1)
    #     # exclude non-active
    #     if only_running and active != 'active':
    #         continue
    #     if active == 'active':
    #         status = sub
    #     else:
    #         status = active
    #     yield (node, status)

    if not node_filter:
        return

    guard_units = set()

    # find all nodes that systemd knows about.
    proc = systemctl_cmd('list-units', '--all', '--full', '--no-legend', '--plain', '--no-pager',
                         nodes='*',
                         stdout=subprocess.PIPE, universal_newlines=True, bufsize=1,
                         )
    for line in proc.stdout.split('\n'):
        if not line:
            continue
        unit,__ = line.split(maxsplit=1)
        guard_units.add(unit)

    # find all enabled units (linked to appropriate target)
    # HACK: we have to resort to a directory listing here because we
    # want to use "enabled" status to be the definitive list of nodes,
    # but there's no way to list enabled template instances :(
    try:
        guard_units |= set(os.listdir(GUARD_TARGET_PATH))
    except FileNotFoundError:
        pass

    # extract node names from unit names
    all_nodes = set()
    for unit in guard_units:
        node_match = UNIT2NODE_RE.match(unit)
        if not node_match:
            continue
        node = node_match.group(1)
        if not node:
            continue
        all_nodes.add(node)

    if not all_nodes:
        return

    # filter list based on supplied patterns
    matched_nodes = set()
    for pattern in node_filter:
        match = set(fnmatch.filter(all_nodes, pattern))
        if not match:
            raise util.GuardCtrlError("Pattern '{}' matches no known nodes.".format(pattern))
        matched_nodes |= match

    # check status
    if include_status or only_active:
        for node in sorted(matched_nodes):
            active = systemctl_cmd('is-active', nodes=[node],
                                   check=False, universal_newlines=True,
                                   stdout=subprocess.PIPE).stdout.strip()
            if only_active and active != 'active':
                continue
            if include_status:
                enabled = systemctl_cmd('is-enabled', nodes=[node],
                                        check=False, universal_newlines=True,
                                        stdout=subprocess.PIPE).stdout.strip()
                # FIXME: this should be last state change time
                estl = systemctl_cmd('show', '--property=ExecMainStartTimestamp', nodes=[node],
                                     check=False, universal_newlines=True,
                                     stdout=subprocess.PIPE).stdout.strip()
                execstart = gpstime.parse(estl.split('=', maxsplit=1)[1])
                yield node, enabled, active, execstart
            else:
                yield node

    else:
        for node in sorted(matched_nodes):
            yield node


def print_node_status(nodes=None, only_failed=False, print_logs=True):
    """Print node service status.

    With no nodes specified displays full system status with process
    tree.  If nodes are specified their individual status is displayed
    along with their last few lines in the log.

    """
    nodes = list(list_nodes(nodes))
    args = ['status', '--quiet', '--no-pager']
    if only_failed:
        args += ['--state=failed']
    if not nodes:
        args += ['guardian.slice']
    if not print_logs:
        args += ['--line', '0']
    try:
        systemctl_cmd(*args, nodes=nodes)
    except SystemctlError:
        pass


def enable_nodes(nodes, now=False):
    """Enable nodes.

    Enabled nodes are registered by the system and auto-started at
    boot.

    """
    util.test_chanfile()
    for node in nodes:
        util.check_system(node)
    logging.info("enabling nodes: {}".format(' '.join(nodes)))
    cmd = ['enable']
    if now:
        cmd += ['--now']
    try:
        systemctl_cmd(*cmd, nodes=nodes)
    except SystemctlError as e:
        raise util.GuardCtrlError("Failed to enable node: {}".format(e.stderr))
    util.update_global_chanfile(list(list_nodes()))


def start_nodes(nodes):
    """Start nodes.

    Start not running nodes.  Nodes must be enabled before starting.
    Is a no-op for already running nodes.

    """
    nodes = list(list_nodes(nodes))
    if not nodes:
        raise util.GuardCtrlError("Pattern(s) match no enabled node(s).  Please 'enable' before 'start'.")
    logging.info("starting nodes: {}".format(' '.join(nodes)))
    try:
        systemctl_cmd('start', nodes=nodes)
    except SystemctlError:
        print_node_status(nodes=nodes, only_failed=True, print_logs=True)
        # reset failed status and clear node
        # FIXME: is this the right way to handle this?
        # systemctl_cmd('reset-failed', nodes=nodes)
        raise util.GuardCtrlError("Failed to start node(s).")


def restart_nodes(nodes):
    """Restart nodes.

    Restart running nodes, or start nodes that aren't running.  Nodes
    must be enabled before starting.

    """
    nodes = list(list_nodes(nodes))
    if not nodes:
        raise util.GuardCtrlError("Pattern(s) match no enabled node(s).")
    logging.info("restarting nodes: {}".format(' '.join(nodes)))
    try:
        systemctl_cmd('restart', nodes=nodes)
    except SystemctlError:
        print_node_status(nodes=nodes, only_failed=True, print_logs=True)
        # reset failed status and clear node
        # FIXME: is this the right way to handle this?
        # systemctl_cmd('reset-failed', nodes=nodes)
        raise util.GuardCtrlError("Failed to restart node(s).")


def stop_nodes(nodes):
    """Stop nodes.

    Stop running nodes.  No-op for non-running nodes.

    """
    nodes = list(list_nodes(nodes))
    if not nodes:
        raise util.GuardCtrlError("Pattern(s) match no enabled node(s).")
    logging.info("stopping nodes: {}".format(' '.join(nodes)))
    try:
        systemctl_cmd('stop', nodes=nodes)
    except SystemctlError as e:
        print_node_status(nodes=nodes, only_failed=True, print_logs=True)
        raise util.GuardCtrlError("Failed to stop node.")


def disable_nodes(nodes, now=False):
    """Disable nodes.

    Disabling nodes removes them from the site registering.  If the
    node is running it will be stopped.

    """
    util.test_chanfile()
    nodes = list(list_nodes(nodes))
    if not nodes:
        raise util.GuardCtrlError("Pattern(s) match no enabled node(s).")
    logging.info("disabling nodes: {}".format(' '.join(nodes)))
    cmd = ['disable']
    if now:
        cmd += ['--now']
    try:
        systemctl_cmd(*cmd, nodes=nodes)
    except SystemctlError as e:
        print_node_status(nodes=nodes, only_failed=True, print_logs=True)
        raise util.GuardCtrlError("Failed to disable node.")
    systemctl_cmd('reset-failed', nodes=nodes, check=False, stderr=subprocess.PIPE)
    util.update_global_chanfile(list(list_nodes()))

##################################################

def node_logs(nodes=None,
              nlines=None, after=None, before=None,
              follow=False):
    """Generator of node log messages.

    yields (gpstime, message) tuples.

    `after` and `before` options expect 'gpstime' objects.

    """
    cmd = ['journalctl', #'--user',
           '--no-pager', '--quiet', '--no-hostname',
           '--output', 'short-unix',
           ]

    if nlines:
        cmd += ['--lines', str(nlines)]
    if after:
        cmd += ['--since', after.astimezone(tzlocal()).strftime(JOURNALCTL_TIME_FMT)]
    if before:
        cmd += ['--until', before.astimezone(tzlocal()).strftime(JOURNALCTL_TIME_FMT)]
    if follow:
        cmd += ['--follow']
    for node in nodes:
        cmd += ['--user-unit={}'.format(UNIT_FMT.format(node))]
    logging.debug('cmd: ' + ' '.join(cmd))

    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL,
                            universal_newlines=True, bufsize=1)
    for line in proc.stdout:
        line = line.rstrip()
        try:
            time, name, msg = line.split(' ', maxsplit=2)
            dt = gpstime.fromtimestamp(float(time), tz=tzlocal())
        except:
            dt = None
            msg = line
        yield dt, msg



color_palette = {
    'white':            'ffffff',
    'black':            '000000',
    'state_normal':     'b3cde3',
    'state_normal_req': '377eb8',
    'state_goto':       'ccebc5',
    'state_goto_req':   '4daf4a',
    'state_init':       'decbe4',
    'state_init_req':   '984ea3',
    'state_none':       '919191',
    }

medm_default_color_list = [
    'ffffff',
    'ececec',
    'dadada',
    'c8c8c8',
    'bbbbbb',
    'aeaeae',
    '9e9e9e',
    '919191',
    '858585',
    '787878',
    '696969',
    '5a5a5a',
    '464646',
    '2d2d2d',
    '000000',
    '00d800',
    '1ebb00',
    '339900',
    '2d7f00',
    '216c00',
    'fd0000',
    'de1309',
    'be190b',
    'a01207',
    '820400',
    '5893ff',
    '597ee1',
    '4b6ec7',
    '3a5eab',
    '27548d',
    'fbf34a',
    'f9da3c',
    'eeb62b',
    'e19015',
    'cd6100',
    'ffb0ff',
    'd67fe2',
    'ae4ebc',
    '8b1a96',
    '610a75',
    'a4aaff',
    '8793e2',
    '6a73c1',
    '4d52a4',
    '343386',
    'c7bb6d',
    'b79d5c',
    'a47e3c',
    '7d5627',
    '58340f',
    '99ffff',
    '73dfff',
    '4ea5f9',
    '2a63e4',
    '0a00b8',
    'ebf1b5',
    'd4db9d',
    'bbc187',
    'a6a462',
    '8b8239',
    '73ff6b',
    '52da3b',
    '3cb420',
    '289315',
    '1a7309',
]

grd_color_list = [
    'ffffff',
    '000000',
    'c8c8c8',
    'bbbbbb',
    '858585',
    '464646',
    '339900',
    'fd0000',
    '5893ff',
    '597ee1',
    '4b6ec7',
    '3a5eab',
    '27548d',
    'fbf34a',
    'cd6100',
    'ffb0ff',
    '8b1a96',
    '4ea5f9',
]

# import sys
# print color_list[int(sys.argv[1])]

#print len(color_list), color_list[-1]
color_list = medm_default_color_list[:-len(color_palette)-1]
for i, value in enumerate(color_palette.values()):
    ii = i+1
    if value not in color_list:
        color_list.append(value)
#print len(color_list), color_list[-1]

def crslv(name):
    return color_list.index(color_palette[name])

##################################################

def header(**kwargs):
    if 'bclr' not in kwargs:
        kwargs['bclr'] = 29
    ss = '''
file {{
	name="GRD_AUTOGEN.adl"
	version=030107
}}
display {{
	object {{
		x=0
		y=0
		width={width}
		height={height}
	}}
	clr=14
	bclr={bclr}
	cmap=""
	gridSpacing=5
	gridOn=0
	snapToGrid=0
}}
'''.format(**kwargs)
    ss += '''
"color map" {{
	ncolors=65
	colors {{
'''
    for color in color_list:
        ss += '''
               {0},'''.format(color)
    ss += '''
}}
'''
    return ss

def background(**kwargs):
    return '''
rectangle {{
	object {{
		x=0
		y=0
		width={width}
		height={height}
	}}
	"basic attribute" {{
		clr=29
	}}
}}
rectangle {{
	object {{
		x=0
		y=0
		width={width}
		height={height}
	}}
	"basic attribute" {{
		clr=39
	}}
	"dynamic attribute" {{
		vis="calc"
		calc="A=1"
		chan="$(IFO):GRD-$(SYSTEM)_MODE"
	}}
}}
rectangle {{
	object {{
		x=0
		y=0
		width={width}
		height={height}
	}}
	"basic attribute" {{
		clr=35
	}}
	"dynamic attribute" {{
		vis="calc"
		calc="A=2"
		chan="$(IFO):GRD-$(SYSTEM)_MODE"
	}}
}}
rectangle {{
	object {{
		x=0
		y=0
		width={width}
		height={height}
	}}
	"basic attribute" {{
		clr=24
	}}
	"dynamic attribute" {{
		vis="if not zero"
		chan="$(IFO):GRD-$(SYSTEM)_CONNECT"
	}}
}}
rectangle {{
	object {{
		x=0
		y=0
		width={width}
		height={height}
	}}
	"basic attribute" {{
		clr=30
	}}
	"dynamic attribute" {{
		vis="calc"
		calc="A=1"
		chan="$(IFO):GRD-$(SYSTEM)_OP"
	}}
}}
rectangle {{
	object {{
		x=0
		y=0
		width={width}
		height={height}
	}}
	"basic attribute" {{
		clr=20
	}}
	"dynamic attribute" {{
		vis="if not zero"
		chan="$(IFO):GRD-$(SYSTEM)_ERROR"
	}}
}}
rectangle {{
	object {{
		x=0
		y=0
		width={width}
		height=18
	}}
	"basic attribute" {{
		clr=29
	}}
	"dynamic attribute" {{
		chan="BLAH"
	}}
}}
polyline {{
	object {{
		x=0
		y=17
		width={width}
		height=1
	}}
	"basic attribute" {{
		clr=25
		fill="outline"
		width=1
	}}
	"dynamic attribute" {{
		chan="BLAH"
	}}
	points {{
		(0,17)
		({width},17)
	}}
}}
'''.format(**kwargs)

def title(x):
    return '''
text {{
	object {{
		x={x}
		y=2
		width=100
		height=12
	}}
	"basic attribute" {{
		clr=4
	}}
	"dynamic attribute" {{
		chan="BLAH"
	}}
	textix="GUARDIAN:"
	align="horiz. right"
}}
text {{
	object {{
		x={x1}
		y=2
		width=100
		height=12
	}}
	"basic attribute" {{
		clr=0
	}}
	"dynamic attribute" {{
		chan="BLAH"
	}}
	textix="$(SYSTEM)"
	align="horiz. left"
}}
'''.format(x=x, x1=x+102)

##################################################

def related_display_menu(displays, **kwargs):
    ss = '''
"related display" {{
       object {{
               x={x}
               y={y}
               width={width}
               height={height}
       }}'''.format(**kwargs)
    for i, display in enumerate(displays):
        ss +='''
       display[{0}] {{
               label="{1}"
               name="{2}"
               args="{3}"
       }}'''.format(i, display[0], display[1], display[2])
    ss += '''
       clr=0
       bclr=12
       label="related"
}}
'''
    return ss

##################################################

def state_request_button(**kwargs):
    if 'width' not in kwargs:
        kwargs['width'] = 270

    kwargs['xe'] = kwargs['x']
    kwargs['widthe'] = 18
    kwargs['xb'] = kwargs['x'] + kwargs['widthe']
    kwargs['widthb'] = kwargs['width'] - kwargs['widthe']

    state = kwargs['state']
    is_goto = kwargs['is_goto']
    is_request = kwargs['is_request']

    clr = crslv('black')
    bclr = crslv('state_normal')

    if state == 'NONE':
        bclr = crslv('state_none')
    elif state == 'INIT':
        if is_request:
            clr = crslv('white')
            bclr = crslv('state_init_req')
        else:
            bclr = crslv('state_init')
    elif is_goto:
        if is_request:
            clr = crslv('white')
            bclr = crslv('state_goto_req')
        else:
            bclr = crslv('state_goto')
    else:
        if is_request:
            clr = crslv('white')
            bclr = crslv('state_normal_req')
    kwargs['clr'] = clr
    kwargs['bclr'] = bclr

    return '''
"shell command" {{
	object {{
		x={xe}
		y={y}
		width={widthe}
		height=26
	}}
	command[0] {{
		label="edit"
		name="guardutil"
		args="edit $(SYSTEM) {state} &"
	}}
	clr=0
	bclr=12
	label="-e"
}}
"message button" {{
	object {{
		x={xb}
		y={y}
		width={widthb}
		height=26
	}}
	control {{
		chan="$(IFO):GRD-$(SYSTEM)_REQUEST"
		clr={clr}
		bclr={bclr}
	}}
	label="{state}"
        release_msg="{state}"
}}
'''.format(**kwargs)
